//#include <iostream>
//#include <ios>
//#include <fstream>
//#include <map>
//#include <algorithm>
//
//#include <boost/filesystem.hpp>
//#include <zeep/xml/document.hpp>
//#include <boost/foreach.hpp>
//#define foreach BOOST_FOREACH
//#include <boost/algorithm/string.hpp>
//#include <boost/format.hpp>

#define BOOST_TEST_MAIN
#include <boost/test/unit_test.hpp>


