* {
	font-family: sans-serif;
	font-size: small;
}

html, body {
	padding: 0px;
	margin: 0px;
	background: #fff;
}

#main {
	padding: 4px;
	margin: 0;
	font-size: small;
}

@media all and (min-width: 800px)
{
	#main { padding: 8px; }
}

#footer {
	position: absolute;
	bottom: 0px;
	height: 1em;
	left: 0px;
	right: 0px;
	overflow: hidden;
	padding: 4px;
	margin: 0;
	font-size: x-small;
	background: #f0f0f0;
}

#footer * {
	font-size: x-small;
}

@media all and (min-width: 800px)
{
	#footer { padding: 8px; }
}



/* lists */
table.list {
	padding: 0px;
	margin: 0px;
	vertical-align: top;
}

@media all and (min-width: 800px)
{
	table.list {
		min-width: 450px;
	}
}

.list caption {
	background-color: #449d9d;
	color: white;
	font-weight: bold;
	text-align: left;
	padding: 3px 5px 3px 5px;
}

.list th {
	background-color: #c6e0e1;
	color: #2f6c6c;
	border-bottom: solid 1px #449d9d;
	font-weight: bold;
	text-align: left;
	text-decoration: none;
	padding: 2px 5px 2px 5px;
}

.list th + th {
	border-left: 1px solid #449d9d;
}

.list th a {
	text-decoration: underline;
}

.list td {
	padding: 2px;
	padding-left: 4px;
	margin: 0px;
	vertical-align: top;
	border-top: solid 1px #ebe7e0;
}

table.list tr:first-child td {
	border-top-width: 0px;
}

.list .row-title {
	display: block;
}

td.right {
	text-align: right;
}

.list td.nr {
	font-weight: bold;
	text-align: right;
}

.list a img {
	border: none 0px;
}

.head {
	display: block;
	height: 5em;
	line-height: 5em;
	background-color: #6f1919;
	color: white;
	font-size: normal;
	font-weight: bold;
}

.head span {
	display: inline-block;
	vertical-align: middle;
	line-height: normal;
}

.head .error-nr {
	font-size: xx-large;
	padding: 0 0.5em 0 0.5em;
}

.error-head-text {
	font-size: large;
}
