// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently, but
// are changed infrequently
//

#pragma once

#include "targetver.h"

#include <stdio.h>
#include <tchar.h>

#define not !
#define and &&
#define or ||

// the next line is to avoid adding dependancies on the c99l version of the math libraries
#define BOOST_MATH_NO_LONG_DOUBLE_MATH_FUNCTIONS

// TODO: reference additional headers your program requires here
